package com.example.rpserver.controller;

import com.example.rpserver.model.Credential;
import com.example.rpserver.model.FidoUser;
import com.example.rpserver.model.Response;
import com.example.rpserver.repository.CredentialRepository;
import com.example.rpserver.repository.UserRepository;
import com.webauthn4j.WebAuthnManager;
import com.webauthn4j.data.*;
import com.webauthn4j.data.attestation.statement.COSEAlgorithmIdentifier;
import com.webauthn4j.data.client.challenge.Challenge;
import com.webauthn4j.data.client.challenge.DefaultChallenge;
import com.webauthn4j.util.Base64Util;
import com.webauthn4j.util.HexUtil;
import com.webauthn4j.validator.attestation.statement.AttestationStatementValidator;
import com.webauthn4j.validator.attestation.trustworthiness.certpath.NullCertPathTrustworthinessValidator;
import com.webauthn4j.validator.attestation.trustworthiness.self.NullSelfAttestationTrustworthinessValidator;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.HttpHeaders;
import org.springframework.http.ResponseEntity;
import org.springframework.web.bind.annotation.*;

import java.security.SecureRandom;
import java.util.Date;
import java.util.LinkedList;

@RestController
public class MakeCredential {
	private final Logger logger = LoggerFactory.getLogger(MakeCredential.class);
	private final SecureRandom secureRandom = new SecureRandom();

	@Autowired
	private UserRepository userRepository;

	@Autowired
	private CredentialRepository credentialRepository;

	@ResponseBody
	@PostMapping("/BeginMakeCredential")
	public ResponseEntity<PublicKeyCredentialCreationOptions> beginMakeCredential(@RequestParam("username") String username,
	                                                                              @RequestParam("advanced") String advanced,
	                                                                              @RequestParam("advancedOptions") String advancedOptions,
	                                                                              @RequestHeader(HttpHeaders.HOST) String host){
		byte[] userIdBytes = new byte[32];
		secureRandom.nextBytes(userIdBytes);
		PublicKeyCredentialUserEntity userEntity = new PublicKeyCredentialUserEntity(userIdBytes, username, username);
		PublicKeyCredentialRpEntity rp = new PublicKeyCredentialRpEntity("localhost", "localhost");
		Challenge challenge = new DefaultChallenge();
		LinkedList<PublicKeyCredentialParameters> parameters = new LinkedList<>();
		parameters.add(new PublicKeyCredentialParameters(PublicKeyCredentialType.PUBLIC_KEY, COSEAlgorithmIdentifier.RS256));
		parameters.add(new PublicKeyCredentialParameters(PublicKeyCredentialType.PUBLIC_KEY, COSEAlgorithmIdentifier.ES256));
		PublicKeyCredentialCreationOptions options = new PublicKeyCredentialCreationOptions(rp, userEntity, challenge, parameters);

		FidoUser fidoUser = new FidoUser(userIdBytes, challenge.toString(), username);
		userRepository.save(fidoUser);

		return ResponseEntity.ok(options);
	}

	@ResponseBody
	@PostMapping("/FinishMakeCredential")
	public ResponseEntity<Response> finishMakeCredential(@RequestParam("identifier") String identifier,
														 @RequestParam("rawId") byte[] rawId,
	                                                     @RequestParam("clientDataJSON") byte[] clientDataJSONBytes,
	                                                     @RequestParam("attestationObject") byte[] attestationObjectBytes){
		// apply base64 decoding
		clientDataJSONBytes = Base64Util.decode(clientDataJSONBytes);
		attestationObjectBytes = Base64Util.decode(attestationObjectBytes);
		rawId = Base64Util.decode(rawId);

		// logger
		logger.info("clientDataJson : {}", new String(clientDataJSONBytes));
		logger.info("attestationObject hex String : {}", HexUtil.encodeToString(attestationObjectBytes));
		logger.info("rawId hex String : {}", HexUtil.encodeToString(rawId));

		RegistrationRequest registrationRequest = new RegistrationRequest(attestationObjectBytes, clientDataJSONBytes);

		LinkedList<AttestationStatementValidator> attestationStatementValidators = new LinkedList<>();
		WebAuthnManager webAuthnManager = new WebAuthnManager(attestationStatementValidators, new NullCertPathTrustworthinessValidator(), new NullSelfAttestationTrustworthinessValidator());
		RegistrationData registrationData = webAuthnManager.parse(registrationRequest);
		logger.info(registrationData.toString());

		FidoUser user = userRepository.findByChallenge(registrationData.getCollectedClientData().getChallenge().toString());

		credentialRepository.save(new Credential(
				new Date(),
				registrationData.getAttestationObject().getAuthenticatorData().getSignCount(),
				identifier,
				null,
				user.getUserName(),
				rawId));

		return ResponseEntity.ok().body(new Response(true, "Successfully created credential"));
	}
}
